import numpy as N
import matplotlib.pyplot as P


# Mandelbrot algorithm: z[i+1] = z[i]^2 + c, z[0] = 0
#                       with c = [-2,1] x [-1,1]
#                       escape if |z| > 2


# here's the internal loop in Python, using Boolean-indexed numpy arrays,
# and defining the complex-plane grid c as a complex numpy array
# note that we're using complex arithmetics

def pymandel(esc,maxi):
    ny, nx = esc.shape                      # grab the size of the array
    
    x,y = N.meshgrid(N.linspace(-2.0,1.0,nx),N.linspace(-1.0,1.0,ny))
    c = x + 1j*y                            # build a grid of complex coordinates
                                            # (note second matrix dimension is x axis)
    
    z = N.zeros((ny,nx),'complex128')
    
    esc[:,:] = maxi                         # set all points to the maximum iteration count
    
    for i in xrange(1,maxi):
        mask = (esc == maxi)                # get a Boolean "mask" to work only on non-escaped points
        
        z[mask] = z[mask]**2 + c[mask]      # Mandelbrot iteration
        
        esc[mask & (N.abs(z) > 2)] = i      # assign iteration number color to just-escaped points
                                            # (& is logical and; we must be careful about operator precedence)
    
    return esc


# here's the Python driver... it allocates memory for the Mandelbrot iteration
# count, calls mandelloop (which may be Python, or a wrapped C or Fortran function)
# to get it, then plots it with matplotlib.pyplot.imshow()

def mandelplot(n=100,maxi=256,plot=True,mandelloop=pymandel,arrayorder='C'):
    nx,ny = (3*n)/2,n
    
    esc = N.zeros((ny,nx),'int32',order=arrayorder)
    
    mandelloop(esc,maxi)
    
    if plot:
        P.imshow(esc,origin='lower',extent=(-2.0,1.0,-1.0,1.0))
    
    return esc


# the C internal loop is defined in cmandel.c, and wrapped with cython in pyxmandel.pyx,
# resulting in pyxmandel.so (run "make pyxmandel.so" to make it)

# the same C code is wrapped with SWIG in swigmandel.i (run "make _swigmandel.so")

# finally, the Fortran internal loop is defined in fmandel.f90 and wrapped with f2py,
# resulting in fmandel.so (run "make fmandel.so")

# let's check performance in ipython:
# >>> import lecture3 as l3
# >>> %time l3.mandelplot(400,mandelloop=l3.pymandel);    # suppress printout of array
# CPU times: user 4.58 s, sys: 0.35 s, total: 4.93 s
# Wall time: 4.93 s

# >>> import pyxmandel
# >>> %time l3.mandelplot(400,mandelloop=pyxmandel.pyxmandel);
# CPU times: user 0.13 s, sys: 0.00 s, total: 0.13 s
# Wall time: 0.13 s

# >>> import swigmandel
# >>> %time l3.mandelplot(400,mandelloop=swigmandel.cmandel);
# CPU times: user 0.13 s, sys: 0.00 s, total: 0.13 s
# Wall time: 0.13 s

# >>> import fmandel
# >>> %time l3.mandelplot(400,mandelloop=fmandel.fmand,arrayorder='F');
# CPU times: user 0.43 s, sys: 0.00 s, total: 0.43 s
# Wall time: 0.43 s

# so C is 38 times faster than Python, and ~3 times faster than Fortran
# (which is a little unexpected, but maybe caused by compilers)
